package main

import (
	"compress/gzip"
	"bufio"
	"flag"
	"fmt"
	"io"
	"json"
	"os"
	"syscall"
	"time"
)


func zcat_lines(in io.Reader, out chan string) (err os.Error) {
	defer func () {
		if r := recover(); r != nil {
			os.Stderr.Write([]byte("Panic in zcat_lines\n"))
		}
		close(out)
	}()
	buf_in, err := bufio.NewReaderSize(in, 65536)
	if err != nil {
		return
	}
	z_in, err := gzip.NewReader(buf_in)
	defer z_in.Close()
	if err != nil {
		return
	}
	buf_z, err := bufio.NewReaderSize(z_in, 65536)
	if err != nil {
		return
	}
	prefix := ""
	result := ""
	for {
		line, isPrefix, err := buf_z.ReadLine()
		if isPrefix {
			prefix = prefix + string(line)
			continue
		}
		if err == os.EOF {
			break
		}
		if err != nil {
			fmt.Fprintf(os.Stderr, "Error reading line: %s\n", err)
			break
		}
		result = string(line)
		if len(prefix) > 0 {
			result = prefix + result
			// fmt.Fprintf(os.Stderr, "\nlong line: %d\n", len(result))
		}
		prefix = ""
		out <- result
	}
	return
}

type MeliaeMemObject struct {
	Address uint64
	Type string
	Size uint64
	Len uint64
	Refs []uint64
	Value interface{} // can be an Int, String, or Bool
}

func decode_json(in chan string, out chan MeliaeMemObject) (err os.Error) {
	defer func () {
		if r := recover(); r != nil {
			os.Stderr.Write([]byte("Panic in decode_json\n"))
		}
		close(out)
	}()
	for {
		line, ok := <-in
		if !ok {
			break
		}
		// fmt.Fprintf(os.Stderr, "deco: %#v\n", string(line))
		decoded := new(MeliaeMemObject)
		err := json.Unmarshal([]byte(line), decoded)
		if err != nil {
			if len(line) > 200 {
				line = line[:197] + "..."
			}
			fmt.Fprintf(os.Stderr, "Error decoding line: %#v\n%s\n",
				line, err)
			break
		}
		out <- *decoded
	}
	return
}

var keep = flag.Bool("keep", false, "Keep the data in memory while parsing")

func spawn_usage() {
	var args = []string{"grep", "--color", "VmPeak"}
	args = append(args, fmt.Sprintf("/proc/%d/status", os.Getpid()))
	proc_attr := os.ProcAttr{}
	proc_attr.Files = []*os.File{os.Stdin, os.Stdout, os.Stderr}
	proc, err := os.StartProcess("/bin/grep", args, &proc_attr)
	if err != nil {
		os.Stderr.Write([]byte("Failed to spawn\n"))
		return
	}
	proc.Wait(0)
}

func set_limits() {
// 	gb25 := uint64(2.5 * 1024 * 1024 * 1024)
// 	// gccgo doesn't seem to include the zsyscall_linux_amd64* functionality.
// 	// It seems to expose the constants like SYS_SETRLIMIT, but none of the
// 	// functionality (Setrlimit nor RawSyscall)
// 	// I think it is just an oversight, but for now we have to skip this
// 	rlimit := Rlimit{gb25, gb25}
// 	// Why isn't this in syscall?
// 	const RLIMIT_AS = 9
// 	// GCCGO doesn't expose Setrlimit, nor does it expose RawSyscall
// 	// syscall.Setrlimit(RLIMIT_AS, &rlimit)
}

func main() {
	flag.Parse()
	var in_f *os.File
	var err os.Error

	if flag.NArg() >= 1 {
		fname := flag.Arg(0)
		in_f, err = os.Open(fname)
		if err != nil {
			fmt.Fprintf(os.Stderr, "Failed to open input file: \"%s\"\n%s\n",
				fname, err)
			return
		}
		defer in_f.Close()
	} else {
		in_f = os.Stdin
	}
	if flag.NArg() > 1 {
		fmt.Fprintf(os.Stderr,
			"Too many arguments, did you mean to set a flag?\n"+
				"It must come before the arguments.\n")
		return
	}
	set_limits()
	start := time.Nanoseconds()
	line_stream := make(chan string, 100)
	go zcat_lines(in_f, line_stream)
	decoded_ch := make(chan MeliaeMemObject, 100)
	go decode_json(line_stream, decoded_ch)
	count := 0
	last_read_lines := []byte("")
	defer func() {
		if r := recover(); r != nil {
			os.Stdout.Write(last_read_lines)
			os.Stdout.Write([]byte{'\n'})
			// We can't complete because these funcs want to allocate memory.
			return
		}
		delta := time.Nanoseconds() - start
		var rusage syscall.Rusage
		syscall.Getrusage(0, &rusage)
		fmt.Printf("Read %d lines in %.3fs\n", count, float64(delta)*1e-9)
		fmt.Printf("Peak Mem: %.1fMiB\n", float64(rusage.Maxrss)/1024.)
		spawn_usage()
	}()
	if *keep {
		fmt.Printf("Keeping references\n")
	}
	all_res := make([]MeliaeMemObject, 1000)
	for {
		val, ok := <-decoded_ch
		if !ok {
			// we are done
			break
		}
		if *keep {
			all_res = append(all_res, val)
		}
		count += 1
		if count%10000 == 0 {
			delta := time.Nanoseconds() - start
			read := fmt.Sprintf("Read %d lines in %.3fs\r",
				count, float64(delta)*1e-9)
			last_read_lines = []byte(read)
			os.Stdout.Write(last_read_lines)
		}
	}
	return
}
